import sympy

import types
import sys

from functools import wraps


class Eq(sympy.Eq):

    @property
    def rhs(self):
        return super().rhs

    @rhs.setter
    def rhs(self, rhs):
        self._args = (self.lhs, rhs)

    @property
    def lhs(self):
        return super().lhs

    @lhs.setter
    def lhs(self, lhs):
        self._args = (lhs, self.rhs)

    def __add__(self, other):
        if isinstance(other, Eq):
            return Eq(self.lhs + other.lhs, self.rhs + other.rhs)
        return Eq(self.lhs + other, self.rhs + other)

    def __radd__(self, other):
        return Eq(self.lhs + other, self.rhs + other)

    def __sub__(self, other):
        if isinstance(other, Eq):
            return Eq(self.lhs - other.lhs, self.rhs - other.rhs)
        return Eq(self.lhs - other, self.rhs - other)

    def __rsub__(self, other):
        return Eq(self.lhs - other, self.rhs - other)

    def __mul__(self, other):
        if isinstance(other, Eq):
            return Eq(self.lhs * other.lhs, self.rhs * other.rhs)
        return Eq(self.lhs * other, self.rhs * other)

    def __rmul__(self, other):
        return Eq(other * self.lhs, other * self.rhs)

    def __truediv__(self, other):
        if isinstance(other, Eq):
            return Eq(self.lhs / other.lhs, self.rhs / other.rhs)
        return Eq(self.lhs / other, self.rhs / other)

    def __rtruediv__(self, other):
        return Eq(other / self.lhs, other / self.rhs)

    def __pow__(self, power, modulo=None):
        return Eq(self.lhs ** power, self.rhs ** power)

    def apply(self, side, what, *args, **kwargs):
        if side == 'lhs':
            return Eq(what(self.lhs, *args, **kwargs), self.rhs)
        elif side == 'rhs':
            return Eq(self.lhs, what(self.rhs, *args, **kwargs))
        elif side == 'both':
            return Eq(what(self.lhs, *args, **kwargs), what(self.rhs, *args, **kwargs))
        else:
            raise Exception('Unknown key word: %s' % side)

    def doit(self, **kwargs):
        return Eq(self.lhs.doit(**kwargs), self.rhs.doit(**kwargs))


this_module = sys.modules[__name__]


def _get_imported_names(module):
    names = module.__all__ if hasattr(module, '__all__') else dir(module)
    return [name for name in names if not name.startswith('_')]


def _wrap_function(func):
    @wraps(func)
    def f(*args, **kwargs):
        if isinstance(args[0], Eq):
            if len(args) > 1:
                other_args = tuple(args[1:])
            else:
                other_args = ()
            return args[0].apply('both', func, *other_args, **kwargs)
        else:
            return func(*args, **kwargs)

    return f


_names_from_sympy = _get_imported_names(sympy)

for name in _names_from_sympy:
    obj = getattr(sympy, name)
    if isinstance(obj, types.FunctionType) or isinstance(obj, sympy.FunctionClass):
        setattr(this_module, name, _wrap_function(obj))



import unittest

from mathcube import Table


class TestTable(unittest.TestCase):

    def test_insert_column(self):

        data = [[1, 2, 3, 4]]
        t = Table(data)

        self.assertEqual((1, 4), t.shape)

        t.insert_column(1)
        self.assertEqual('', t[0, 1])

        self.assertEqual((1, 5), t.shape)

        with self.assertRaises(ValueError):
            t.insert_column(-1)

        t.insert_column()
        self.assertEqual((1, 6), t.shape)
        self.assertEqual('', t[0, 5])

    def test_insert_row(self):

        data = [[1, 2, 3, 4]]
        t = Table(data)

        t.insert_row(0)
        self.assertEqual('', t[0, 1])

        self.assertEqual((2, 4), t.shape)

        with self.assertRaises(ValueError):
            t.insert_row(-1)

        t.insert_row()
        self.assertEqual((3, 4), t.shape)
        self.assertEqual('', t[2, 1])




from mathcube import Table


class HornerMethod:
    """A step by step implementation of Horner's method to evaluate polynomials."""

    def __init__(self, poly, x0):
        """Initialize Horner method to evaluate polynomial at given point.

        Parameters
        ----------
        poly:  sympy polynomial
            The polynomial to evaluate.
        x0:  number
            The point where to evaluate the polynomial.

        Examples
        --------

        The following example creates the Horner table for the polynomial

            .. math::
                3x^3 + 2x^2 - 4x + 7

        at x = 2.

        >>> from sympy import *
        >>> from sympy.abc import x
        >>> import mathcube.stepbystep as sbs
        >>> p = poly(3*x**3+2*x**2-4*x+7, x)
        >>> horner = sbs.HornerMethod(p, 2)
        >>> horner.step()
        >>> horner.step()
        >>> ...

        """
        self.x0 = x0
        t = Table([poly.all_coeffs()])
        t.insert_column(0)
        t.insert_row()
        t.insert_row()
        t[1, 0] = "x=%d" % x0
        t[1, 1] = 0
        t.style[0, :] = 'border-bottom: 1px solid #000;'
        t.style[:, 0] = 'border-right: 1px solid #000;'
        self.table = t
        self._step_no = 0

    def _repr_html_(self):
        return self.table._repr_html_()

    def step(self):

        t = self.table
        _, ncols = t.shape
        if self._step_no == 2 * ncols - 3:
            return self

        def add_step():
            col = self._step_no // 2 + 1
            t[2, col] = t[0, col] + t[1, col]

        def mul_step():
            col = self._step_no // 2 + 2
            t[1, col] = self.x0 * t[2, col - 1]

        if self._step_no % 2 == 0:
            add_step()
        else:
            mul_step()

        self._step_no += 1

        return self



from IPython.display import display as ipython_display
from itertools import product

class Latex:
    delimiter = '$$'

    def __init__(self, latex):
        self._latex = latex

    @property
    def latex(self):
        return self._latex

    def _repr_latex_(self):
        return ' {} {} {} '.format(
            self.delimiter, self.latex, self.delimiter
        )

    def __add__(self, other):
        if isinstance(other, str):
            other = Latex(other)
        return Latex(self.latex + ' ' + other.latex)

    def __radd__(self, other):
        return Latex(other + ' ' + self.latex)


class LatexInline(Latex):
    delimiter = '$'


class Text(Latex):
    delimiter = '$'

    def __init__(self, text):
        self.latex = r'\mbox{%s}' % text


class HTML:
    ...


class Table(HTML):

    def __init__(self, data=None):
        if data is None:
            data = [[]]
        self._data = data
        self.style = CellStyler(self)

    def _repr_html_(self):
        s = '<table>\n'
        for irow, row in enumerate(self._data):
            s += '<tr>\n'
            for icol, col in enumerate(row):
                s += '<td style="%s"> $' % (self.style[irow, icol]) + str(col) + ' $ </td>\n'
            s += '</tr>\n'
        s += '</table>\n'
        return s

    @property
    def shape(self):
        return len(self._data), len(self._data[0])

    def __getitem__(self, key):
        row_idx, col_idx = key
        return self._data[row_idx][col_idx]

    def __setitem__(self, key, value):
        row, col = key
        self._data[row][col] = value

    def insert_column(self, idx=None):
        if idx is None:
            for row in self._data:
                row.append('')
        elif idx < 0:
            raise ValueError('Negative indexing not allowed')
        else:
            for row in self._data:
                row.insert(idx, '')

    def insert_row(self, idx=None):
        nrows, ncols = self.shape
        if idx is None:
            self._data.append(['']*ncols)
        elif idx < 0:
            raise ValueError('Negative indexing not allowed')
        else:
            self._data.insert(idx, ['']*ncols)

    def add_row(self, row_data):
        self._data.append(row_data)


class CellStyler:

    def __init__(self, table):
        self.table = table
        self.styles = {}

    def __setitem__(self, key, value):
        nrows, ncols = self.table.shape
        all_row_idxs = list(range(nrows))
        all_col_idxs = list(range(ncols))
        row_idx, col_idx = key
        if isinstance(row_idx, slice):
            row_idxs = all_row_idxs[row_idx]
        else:
            row_idxs = [row_idx]
        if isinstance(col_idx, slice):
            col_idxs = all_col_idxs[col_idx]
        else:
            col_idxs = [col_idx]

        cell_idxs = product(row_idxs, col_idxs)
        for cell_idx in cell_idxs:
            self.styles[cell_idx] = value

    def __getitem__(self, key):
        return self.styles.get(key, '')

class HashableSlice:

    def __init__(self, slc):
        self.start = slc.start
        self.stop = slc.stop
        self.step = slc.step

    def to_slice(self):
        return slice(self.start, self.stop, self.step)

    def __hash__(self):
        return hash(repr('%s, %s, %s' % (self.start, self.stop, self.step)))


class Output:

    def __init__(self):
        self.lines = []

    def _to_latex(self, text):
        if isinstance(text, str):
            return Latex(text)
        return text

    def append(self, text):
        self.lines.append(self._to_latex(text))

    def _repr_latex_(self):
        out = ''
        for line in self.lines:
            out += line._repr_latex_()
        return out


def display(*args, **kwargs):
    for arg in args:
        if isinstance(arg, list) or isinstance(arg, tuple):
            for item in arg:
                ipython_display(item)
        return
    ipython_display(*args)



import matplotlib as mpl
import matplotlib.path as mpath
import matplotlib.pyplot as plt
import numpy as np
from IPython.core.display_functions import display
from matplotlib.artist import Artist
from matplotlib.patches import Arrow, FancyArrow, Arc, Rectangle
from numpy import pi

ORIGIN = np.zeros(2)
UP = np.array([0, 1])
DOWN = -UP
RIGHT = np.array([1, 0])
LEFT = -RIGHT


class Drawing:

    def __init__(self):
        self.fig = plt.figure(figsize=(12, 12))
        ax = plt.axes()
        ax.plot()
        ax.set_axis_off()
        ax.set_aspect('equal')
        self.ax = ax

    def add(self, *shapes, plane=1):
        for shape in shapes:
            shape.add_to_axes(self.ax, plane)

    def remove(self, *shapes):
        for shape in shapes:
            try:
                shape.remove()
            except:
                pass  # already removed

    def show(self):
        return display(self.fig)


class Group:

    def __init__(self, *shapes):
        self.shapes = list(shapes)

    def add(self, *shapes):
        self.shapes.extend(list(shapes))

    def add_to_axes(self, ax, plane=1):
        for shape in self.shapes:
            shape.add_to_axes(ax, plane)

    def remove(self):
        for shape in self.shapes:
            shape.remove()


class PatchMixin:

    def add_to_axes(self, ax, plane=1):
        self.patch.set_zorder = plane
        ax.add_patch(self.patch)

    def remove(self):
        self.patch.remove()


class TextMixin:

    def add_to_axes(self, ax, plane=1):
        self.patch = ax.text(*self.xy, self.text, **self.kwargs)
        self.patch.set_zorder = plane

    def remove(self):
        Artist.remove(self.patch)


class Shape:
    linewidth = 0.001


class Point(Shape, PatchMixin):
    size = 0.03
    color = 'k'

    def __init__(self, xy, **kwargs):
        self.xy = np.array(xy)
        if 'color' not in kwargs:
            kwargs['color'] = Point.color
        self.kwargs = kwargs
        self.patch = mpl.patches.Circle(self.xy, Point.size, **self.kwargs)


class Circle(Shape, PatchMixin):

    def __init__(self, xy, radius=1):
        super().__init__()
        self.xy = xy
        self.radius = radius
        self.patch = mpl.patches.Circle(self.xy, self.radius, fill=False, color='k')


class Line(Shape, PatchMixin):

    def __init__(self, start, end, **kwargs):
        self.kwargs = kwargs
        self.start = start
        self.end = end
        self.patch = self._make_patch()

    def _make_patch(self):
        Path = mpath.Path
        path_data = [
            (Path.MOVETO, self.start),
            (Path.LINETO, self.end),
        ]
        codes, verts = zip(*path_data)
        path = mpath.Path(verts, codes)
        return mpl.patches.PathPatch(path, **self.kwargs)


class Vector(Shape, PatchMixin):
    linewidth = 0.005
    head_width = 0.035
    head_length = 0.075
    color = 'k'

    def __init__(self, start, end, **kwargs):

        self.start = start
        self.end = end
        dxy = self.end - self.start
        if 'width' not in kwargs:
            kwargs['width'] = Vector.linewidth
        if 'color' not in kwargs:
            kwargs['color'] = Vector.color
        # kwargs['mutation_scale'] = 100
        kwargs['length_includes_head'] = True
        kwargs['head_width'] = Vector.head_width
        kwargs['head_length'] = Vector.head_length
        self.kwargs = kwargs

        self.patch = FancyArrow(*self.start, *dxy, **kwargs)


class Text(Shape, TextMixin):
    fontsize = 14

    def __init__(self, text, xy, **kwargs):
        self.text = text
        self.xy = xy
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = Text.fontsize
        if 'horizontalalignment' not in kwargs:
            kwargs['horizontalalignment'] = 'center'
        if 'verticalalignment' not in kwargs:
            kwargs['verticalalignment'] = 'center'
        self.kwargs = kwargs


class AngleMarker(Shape, PatchMixin):

    def __init__(self, xy, radius, angles=None, between=None, **kwargs):
        if angles:
            start_angle_deg, end_angle_deg = angles
        elif between:
            v1, v2 = between
            angle = angle_between(v1, v2)
            start_angle_deg = get_polar(v1)
            end_angle_deg = start_angle_deg + angle
        self.patch = Arc(
            xy, 2 * radius, 2 * radius, theta1=start_angle_deg, theta2=end_angle_deg,
            **kwargs
        )


class RightAngleMarker(Shape, PatchMixin):
    def __init__(self, xy, size, angle, **kwargs):
        self.patch = Rectangle(xy, size, size, angle, fill=False, **kwargs)


def rotate(xy, angle):
    a = deg2rad(angle)
    return np.array([
        np.cos(a) * xy[0] - np.sin(a) * xy[1],
        np.sin(a) * xy[0] + np.cos(a) * xy[1]
    ])


def rad2deg(*rads):
    degs = [rad / pi * 180 for rad in rads]
    if len(degs) == 1:
        return degs[0]
    return tuple(degs)


def deg2rad(*degs):
    rads = [deg * pi / 180. for deg in degs]
    if len(rads) == 1:
        return rads[0]
    return tuple(rads)


def dot(a, b):
    return np.dot(a, b)


def norm(a):
    return np.sqrt(dot(a, a))


def normalize(a):
    return a / norm(a)


def angle_between(v1, v2):
    angle = np.arccos(dot(v1, v2) / norm(v1) / norm(v2))
    if get_polar(v1) > get_polar(v2):
        angle = 2 * pi - angle
    return rad2deg(angle)


def polar(deg=None, rad=None, r=1):
    if deg is not None and rad is not None:
        raise ValueError('Either rad or deg must be given, not both.')
    if deg is None and rad is None:
        raise ValueError('Either rad or deg must be given.')
    if deg is not None:
        rad = deg2rad(deg)
    return r * np.array([np.cos(rad), np.sin(rad)])


def get_polar(v):
    x, y = v
    if x == 0:
        if y > 0:
            angle = pi / 2
        else:
            angle = 3 * pi / 2
    else:
        angle = np.arctan(abs(y / x))
        if x < 0 and y >= 0:
            angle = pi - angle
        elif x < 0 and y < 0:
            angle += pi
        elif x > 0 and y < 0:
            angle = 2 * pi - angle
    return rad2deg(angle)



import unittest

from sympy.abc import *
from sympy import *
from mathcube import *


class TestEquation(unittest.TestCase):

    def test_add_right_to_equation(self):
        eq = Eq(a + b, c)
        actual = eq + d
        self.assertEqual(actual, Eq(a+b+d, c+d))

    def test_add_left_to_equation(self):
        eq = Eq(a + b, c)
        actual = d + eq
        self.assertEqual(actual, Eq(d+a+b, d+c))

    def test_sub_right_to_equation(self):
        eq = Eq(a + b, c)
        actual = eq - d
        self.assertEqual(actual, Eq(a+b-d, c-d))

    def test_sub_left_to_equation(self):
        eq = Eq(a + b, c)
        actual = -d + eq
        self.assertEqual(Eq(-d+a+b, -d+c), actual)

    def test_mul_right_to_equation(self):
        eq = Eq(a + b, c)
        actual = eq * d
        self.assertEqual(Eq((a+b)*d, c*d), actual)

    def test_mul_left_to_equation(self):
        eq = Eq(a + b, c)
        actual = d * eq
        self.assertEqual(Eq(d*(a+b), d*c), actual)

    def test_div_right_to_equation(self):
        eq = Eq(a + b, c)
        actual = eq / d
        self.assertEqual(Eq((a+b)/d, c/d), actual)

    def test_div_left_to_equation(self):
        eq = Eq(a + b, c)
        actual = d / eq
        self.assertEqual(Eq(d/(a+b), d/c), actual)

    def test_add_two_equations(self):
        eq1 = Eq(a, b)
        eq2 = Eq(c, d)
        actual = eq1 + eq2
        expected = Eq(a+c, b+d)
        self.assertEqual(expected, actual)

    def test_sub_two_equations(self):
        eq1 = Eq(a, b)
        eq2 = Eq(c, d)
        actual = eq1 - eq2
        expected = Eq(a-c, b-d)
        self.assertEqual(expected, actual)

    def test_mul_two_equations(self):
        eq1 = Eq(a, b)
        eq2 = Eq(c, d)
        actual = eq1 * eq2
        expected = Eq(a*c, b*d)
        self.assertEqual(expected, actual)

    def test_div_two_equations(self):
        eq1 = Eq(a, b)
        eq2 = Eq(c, d)
        actual = eq1 / eq2
        expected = Eq(a/c, b/d)
        self.assertEqual(expected, actual)

    def test_raise_equation_to_power(self):
        eq = Eq(a, b)
        actual = eq ** 2
        expected = Eq(eq.lhs**2, eq.rhs**2)
        self.assertEqual(expected, actual)

    def test_sqrt_of_equation(self):
        eq = Eq(x, 4)
        actual = sqrt(eq)
        expected = Eq(sqrt(x), 2)
        self.assertEqual(expected, actual)

    def test_atan_of_equation(self):
        eq = Eq(x, 4)
        actual = atan(eq)
        expected = Eq(atan(x), atan(4))
        self.assertEqual(expected, actual)

    def test_apply_rhs_side_only(self):
        eq = Eq((x+y)**2, (a+b)**2)
        actual = eq.apply('rhs', expand)
        expected = Eq((x + y)**2, a**2 + 2*a*b + b**2)
        self.assertEqual(expected, actual)

    def test_apply_lhs_side_only(self):
        eq = Eq((x+y)**2, (a+b)**2)
        actual = eq.apply('lhs', expand)
        expected = Eq(x**2 + 2*x*y + y**2, (a + b)**2)
        self.assertEqual(expected, actual)

    def test_apply_exp_to_both_sides(self):
        eq = Eq(x, 2)
        actual = eq.apply('both', exp)
        expected = Eq(exp(x), exp(2))
        self.assertEqual(expected, actual)

    def test_wrapped_sympy_functions_on_eq(self):
        x, y = symbols('x, y', real=True)
        eq = Eq(x+I, 3+ 2*I*y)
        actual = im(eq)
        expected = Eq(1, 2*y)
        self.assertEqual(expected, actual)

    def test_re_on_equation(self):
        x, y = symbols('x, y', real=True)
        eq = Eq(x, y + 5)
        actual = sqrt(eq)
        expected = Eq(sqrt(x), sqrt(y+5))
        self.assertEqual(expected, actual)

